/*===============================================================================================
 codec_fc.dll
 Ver. 1.0
 Copyright (c), Blazer Studios, 2006.
 Visit http://andreas.blazer.nu for latest version
 Built using play routine by Michael Schwendt http://xmms-fc.sourceforge.net/
 
===============================================================================================*/
#include "FC.h"
#include "MyTypes.h"
#include "logFile.h"
#include <iostream>
#include <stdio.h>
#include "fmod_errors.h"

extern void mixerInit(udword,int,int,uword);
extern void mixerFillBuffer( void*, udword );
FMOD_RESULT F_CALLBACK fcopen(FMOD_CODEC_STATE *codec, FMOD_MODE usermode, FMOD_CREATESOUNDEXINFO *userexinfo);
FMOD_RESULT F_CALLBACK fcclose(FMOD_CODEC_STATE *codec);
FMOD_RESULT F_CALLBACK fcread(FMOD_CODEC_STATE *codec, void *buffer, unsigned int size, unsigned int *read);
FMOD_RESULT F_CALLBACK fcsetposition(FMOD_CODEC_STATE *codec, int subsound, unsigned int position, FMOD_TIMEUNIT postype);

FMOD_CODEC_DESCRIPTION fccodec =
{
    "FMOD Future Composer player plugin",// Name.
    0x00010000,                          // Version 0xAAAABBBB   A = major, B = minor.
    0,                                   // Don't force everything using this codec to be a stream
    FMOD_TIMEUNIT_MS,					 // The time format we would like to accept into setposition/getposition.
    &fcopen,                             // Open callback.
    &fcclose,                            // Close callback.
    &fcread,                             // Read callback.
    0,                                   // Getlength callback.  (If not specified FMOD return the length in FMOD_TIMEUNIT_PCM, FMOD_TIMEUNIT_MS or FMOD_TIMEUNIT_PCMBYTES units based on the lengthpcm member of the FMOD_CODEC structure).
    &fcsetposition,                      // Setposition callback.
    0,                                   // Getposition callback. (only used for timeunit types that are not FMOD_TIMEUNIT_PCM, FMOD_TIMEUNIT_MS and FMOD_TIMEUNIT_PCMBYTES).
    0                                    // Sound create callback (don't need it)
};

class fcplugin
{
	FMOD_CODEC_STATE *_codec;
	
	public:
		fcplugin(FMOD_CODEC_STATE *codec)
		{
			_codec = codec;
			memset(&fcwaveformat, 0, sizeof(fcwaveformat));
		}

		~fcplugin()
		{
			//delete some stuff
		}
	signed short* myBuffer;

	FMOD_CODEC_WAVEFORMAT fcwaveformat;
};
/*
    FMODGetCodecDescription is mandatory for every fmod plugin.  This is the symbol the registerplugin function searches for.
    Must be declared with F_API to make it export as stdcall.
    MUST BE EXTERN'ED AS C!  C++ functions will be mangled incorrectly and not load in fmod.
*/
#ifdef __cplusplus
extern "C" {
#endif

F_DECLSPEC F_DLLEXPORT FMOD_CODEC_DESCRIPTION * F_API FMODGetCodecDescription()
{
    return &fccodec;
}

#ifdef __cplusplus
}
#endif


static int jumpToTime = -1;

FMOD_RESULT F_CALLBACK fcopen(FMOD_CODEC_STATE *codec, FMOD_MODE usermode, FMOD_CREATESOUNDEXINFO *userexinfo)
{


	FMOD_RESULT       result;

	fcplugin *fc = new fcplugin(codec);

	unsigned int bytesread;
	char* smallBuffer;
	smallBuffer = new char[5];

	result = codec->fileseek(codec->filehandle,0,(char*)smallBuffer);
	result = codec->fileread(codec->filehandle,(char*)smallBuffer,5,&bytesread,0);
	
	bool isSMOD = false;
	bool isFC14 = false;
	// Check for "SMOD" ID (Future Composer 1.0 to 1.3).
	isSMOD = (smallBuffer[0]==0x53 && smallBuffer[1]==0x4D &&
              smallBuffer[2]==0x4F && smallBuffer[3]==0x44 &&
              smallBuffer[4]==0x00);
    
    // Check for "FC14" ID (Future Composer 1.4).
    isFC14 = (smallBuffer[0]==0x46 && smallBuffer[1]==0x43 &&
              smallBuffer[2]==0x31 && smallBuffer[3]==0x34 &&
              smallBuffer[4]==0x00);

	delete[] smallBuffer;
	if (!isSMOD && !isFC14)
    {
		return FMOD_ERR_FORMAT;
    }

	
	fc->myBuffer = new signed short[codec->filesize];

	result = codec->fileseek(codec->filehandle,0,(signed short*)fc->myBuffer);
	result = codec->fileread(codec->filehandle,(signed short*)fc->myBuffer,codec->filesize,&bytesread,0);

	FC_init(fc->myBuffer,codec->filesize,0,0);
	mixerInit(44100,16,2,0x80);

	//play through the song reeeeally fast to get the song length
	int msecSongLen = 0;
    do
    {
		FC_play();
        msecSongLen += 20;
    }
    while ( !FC_songEnd );

	fc->fcwaveformat.format       = FMOD_SOUND_FORMAT_PCM16;
	fc->fcwaveformat.channels     = 2;
	fc->fcwaveformat.frequency    = 44100;
	fc->fcwaveformat.blockalign   = (16 >> 3) * fc->fcwaveformat.channels;
	fc->fcwaveformat.lengthpcm    = msecSongLen/1000*fc->fcwaveformat.frequency; //inifite length
	
	codec->waveformat   = &(fc->fcwaveformat);
	codec->numsubsounds = 0;                    /* number of 'subsounds' in this sound.  For most codecs this is 0, only multi sound codecs such as FSB or CDDA have subsounds. */
	codec->plugindata   = fc;                    /* user data value */

	return FMOD_OK;
}

FMOD_RESULT F_CALLBACK fcclose(FMOD_CODEC_STATE *codec)
{
	delete (fcplugin*)codec->plugindata;
	return FMOD_OK;
}

FMOD_RESULT F_CALLBACK fcread(FMOD_CODEC_STATE *codec, void *buffer, unsigned int size, unsigned int *read)
{
	//check if we should jump to a position in the song
	if ( jumpToTime != -1 )
    {
		fcplugin* fc = (fcplugin*)codec->plugindata;
		FC_init(fc->myBuffer,codec->filesize,0,0);
	    while (jumpToTime > 0)
        {
            FC_play();
            jumpToTime -= 20;
        };
        jumpToTime = -1;
	}
	
	mixerFillBuffer(buffer,size);
	*read=size;
	
    return FMOD_OK;
}

FMOD_RESULT F_CALLBACK fcsetposition(FMOD_CODEC_STATE *codec, int subsound, unsigned int position, FMOD_TIMEUNIT postype)
{
	jumpToTime = position;
	return FMOD_OK;
}

